/* 
 *   Klotski: a port of Gnome Klotski to Qt
 *   Gnome Klotski written by Lars Rydlinge <lars.rydlinge@hig.se>
 *   Port to Qt by Philippe Fremy <pfremy@chez.com>
 * 
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "klotski.h"
#include "pieces.h"
#include "levels.h"

void Board::paintEvent( QPaintEvent * pe)
{ emit sg_paintEvent( pe ); }

void Board::mousePressEvent( QMouseEvent * me)
{ emit sg_pressEvent( me ); }

void Board::mouseReleaseEvent( QMouseEvent * me )
{ emit sg_releaseEvent( me ); }

void Board::mouseMoveEvent( QMouseEvent * me )
{ emit sg_moveEvent( me ); }

Klotski::Klotski()
	: QMainWindow()
{
	map = NULL;
	tmpmap = NULL;
	move_map = NULL;
	orig_map = NULL;

	height=-1;
	width=-1;
	moves=0;

	piece_id = -1;
	piece_x = 0; 
	piece_y = 0; 

	buffer = new QPixmap();
	buffer->resize(TILE_SIZE, TILE_SIZE);

	statusbar = new QStatusBar( this );
	move_lab = new QLabel( statusbar );
	set_move( 0 );
	move_lab->setMinimumSize( move_lab->sizeHint() );
	statusbar->addWidget( move_lab, 0, true );

	setCaption( APPNAME );
	init_menu();
	load_image();
	
	space = new Board( this );
	connect( space, SIGNAL( sg_pressEvent(QMouseEvent *) ), 
		this, SLOT( space_mousePressEvent( QMouseEvent *)));
	connect( space, SIGNAL( sg_releaseEvent(QMouseEvent *) ), 
		this, SLOT( space_mouseReleaseEvent( QMouseEvent *)));
	connect( space, SIGNAL( sg_moveEvent(QMouseEvent *) ), 
		this, SLOT( space_mouseMoveEvent( QMouseEvent *)));
	connect( space, SIGNAL( sg_paintEvent(QPaintEvent *) ), 
		this, SLOT( space_paintEvent( QPaintEvent *)));
	setCentralWidget( space );

	new_game();
}

Klotski::~Klotski()
{
	delete tiles_pixmap;
	delete buffer;
}

void Klotski::init_menu() {
	QPopupMenu * level_1_menu = new QPopupMenu( this );
	QPopupMenu * level_2_menu = new QPopupMenu( this );
	QPopupMenu * level_3_menu = new QPopupMenu( this );
	level_1_menu->insertItem( "1", 1);
	level_1_menu->insertItem( "2", 2);
	level_1_menu->insertItem( "3", 3);
	level_1_menu->insertItem( "4", 4);
	level_1_menu->insertItem( "5", 5);
	level_1_menu->insertItem( "6", 6);
	level_1_menu->insertItem( "7", 7);

	level_2_menu->insertItem( "11", 11);
	level_2_menu->insertItem( "12", 12);
	level_2_menu->insertItem( "13", 13);
	level_2_menu->insertItem( "14", 14);
	level_2_menu->insertItem( "15", 15);
	level_2_menu->insertItem( "16", 16);
	level_2_menu->insertItem( "17", 17);

	level_3_menu->insertItem( "21", 21);
	level_3_menu->insertItem( "22", 22);
	level_3_menu->insertItem( "23", 23);
	level_3_menu->insertItem( "24", 24);
	level_3_menu->insertItem( "25", 25);
	level_3_menu->insertItem( "26", 26);

	QPopupMenu * game_menu = new QPopupMenu();
	game_menu->insertItem( "Novice", level_1_menu);
	game_menu->insertItem( "Medium", level_2_menu);
	game_menu->insertItem( "Advanced", level_3_menu);

	QPopupMenu * file_menu = new QPopupMenu();
	file_menu->insertItem( "quit", this, SLOT( quit() ));

	QPopupMenu * help_menu = new QPopupMenu( this );
	help_menu->insertItem( "About klotski", this, SLOT( about() ));
	help_menu->insertItem( "About Qt", this, SLOT( aboutQt() ));

	QMenuBar * main_menu = menuBar();
	main_menu->insertItem( "File", file_menu );
	main_menu->insertItem( "Game", game_menu );
	main_menu->insertItem( "Help", help_menu );

	connect( level_1_menu, SIGNAL( activated( int )),
		this, SLOT( level( int ) ));
	connect( level_2_menu, SIGNAL( activated( int )),
		this, SLOT( level( int ) ));
	connect( level_3_menu, SIGNAL( activated( int )),
		this, SLOT( level( int ) ));

}


void Klotski::space_mousePressEvent(QMouseEvent * event)
{
	piece_x = (int) event->x()/TILE_SIZE;
	piece_y = (int) event->y()/TILE_SIZE;
	piece_id = get_piece_id(map,piece_x,piece_y); 
	copymap(move_map,map);
}


void Klotski::space_mouseReleaseEvent(QMouseEvent * event)
{
	if(movable(piece_id))
		if(mapcmp(move_map,map)) {
			new_move();
			if(game_over()) {
				QMessageBox::information( this, "", "Level completed. Well done.");
			}
		}
}

void Klotski::space_mouseMoveEvent(QMouseEvent * event)
{
	int new_piece_x, new_piece_y;

	new_piece_x = (int) event->x()/TILE_SIZE;
	new_piece_y = (int) event->y()/TILE_SIZE;
	if(new_piece_x >= width || event->x() < 0 ||
	   new_piece_y >= height || event->y() < 0) 
		return;
	if(movable(piece_id))
		if(move_piece(piece_id,piece_x, piece_y, new_piece_x, new_piece_y)==1) {
			piece_x = new_piece_x; piece_y = new_piece_y;
			return;
	}
	return;
}

void Klotski::space_paintEvent(QPaintEvent * pe)
{
	int x,y;
	for(y = 0; y < height; y++)
		for(x = 0; x < width; x++)
			gui_draw_pixmap(map, x, y);
}

void Klotski::gui_draw_pixmap(char *target, int x, int y) 
{

	int value;
	bitBlt( buffer, 0, 0, tiles_pixmap, 
		  get_piece_nr(target,x,y)*TILE_SIZE, 0, 
		  TILE_SIZE, TILE_SIZE, Qt::CopyROP);
  
	if(get_piece_id(target,x,y)=='*'){
		if(get_piece_id(orig_map,x,y)=='.')
			value = 20;
		else
			value = 22;

	bitBlt( buffer, 8, 8, tiles_pixmap, value*TILE_SIZE+10,10,
	    8, 8, Qt::CopyROP);
	}

	QPainter p( space );
	p.drawPixmap( x*TILE_SIZE, y*TILE_SIZE, *buffer ); 
	p.end();
}

void Klotski::quit()
{
	if (moves) {
		if (0 != QMessageBox::warning( this, "Quit", "Really quit ?", "Ok", "Cancel"))
			return;
	}

	qApp->quit();
}

void Klotski::load_image()
{
	tiles_pixmap = new QPixmap( "gnotski.png" );
}

void Klotski::new_game( char * newmap )
{
	prepare_map( newmap );
	space->resize(width*TILE_SIZE, height*TILE_SIZE);
	space->setMinimumSize( space->size() );
	resize( sizeHint() );
	space->repaint();
	set_move(0);
}

int Klotski::movable(int id)
{
	if(!(id == '#' || id == '.' || id == ' ' || id == '-'))
		return 1;
	return 0;
}

void Klotski::set_move(int x) 
{
	moves = x-1;
	new_move();
}

void Klotski::new_move() {
	if(moves<999) moves++;
	QString s;
	s.sprintf("Moves : %4d", moves);
	move_lab->setText( s );
}

void Klotski::print_map(char *src)
{
	int x,y;
	for(y=0; y<height; y++) {
		for(x=0; x<width; x++)
			printf("%c", get_piece_id(src,x,y));
		printf("\n");
	}
}

int Klotski::game_over()
{
	int x,y, over=1;
	for(y=0; y<height; y++)
		for(x=0; x<width; x++)
			if(get_piece_id(map,x,y) == '*' && get_piece_id(orig_map,x,y) != '.')
				over = 0;
	return over;
}

int Klotski::do_move_piece(int id, int dx, int dy)
{
	int x,y;
	copymap(tmpmap,map);

	/* Move pieces */
	for(y=0; y<height; y++)
		for(x=0; x<width; x++)
			if(get_piece_id(tmpmap,x,y) == id)
				set_piece_id(tmpmap,x,y,' ');

	for(y=0; y<height; y++)
		for(x=0; x<width; x++)
			if(get_piece_id(map,x,y) == id)
				set_piece_id(tmpmap,(x+dx),(y+dy),id);

	/* Preserve some from original map */
	for(y=0; y<height; y++)
		for(x=0; x<width; x++){
			if(get_piece_id(tmpmap,x,y)==' ' && get_piece_id(orig_map,x,y)=='.')
				set_piece_id(tmpmap,x,y,'.');
			if(get_piece_id(tmpmap,x,y)==' ' && get_piece_id(orig_map,x,y)=='-')
				set_piece_id(tmpmap,x,y,'-');
		}

	/* Paint changes */
	for(y=0; y<height; y++)
		for(x=0; x<width; x++)
			if(get_piece_id(map,x,y) != get_piece_id(tmpmap,x,y) ||
				 get_piece_id(tmpmap,x,y) == id)
					gui_draw_pixmap(tmpmap, x, y);

	copymap(map,tmpmap);
	return 1;
}

int Klotski::check_valid_move(int id, int dx, int dy)
{
	int x,y,valid = 1;

	for(y=0; y<height; y++)
		for(x=0; x<width; x++)
			if(get_piece_id(map,x,y) == id)
				if(!(get_piece_id(map,x+dx,y+dy) == ' ' ||
				     get_piece_id(map,x+dx,y+dy) == '.' ||
				     get_piece_id(map,x+dx,y+dy) == id ||
				   (id == '*' && get_piece_id(map,x+dx,y+dy) == '-'))){
					valid = 0;
				}
	return valid;
}

int Klotski::move_piece(int id,int x1,int y1,int x2,int y2)
{
	int return_value = 0;

	if(get_piece_id(map,x2,y2) == id)
		return_value = 1;

	if(!((abs(y1-y2)==0 && abs(x1-x2)==1) || (abs(x1-x2)==0 && abs(y1-y2)==1)))
		return 0;

	if(abs(y1-y2)==1){
		if(y1-y2<0) 
			if(check_valid_move(id,0,1))
				return do_move_piece(id,0,1);
		if(y1-y2>0) 
			if(check_valid_move(id,0,-1))
				return do_move_piece(id,0,-1);
	}

	if(abs(x1-x2)==1){
		if(x1-x2<0) 
			if(check_valid_move(id,1,0))
				return do_move_piece(id,1,0);
		if(x1-x2>0) 
			if(check_valid_move(id,-1,0))
				return do_move_piece(id,-1,0);
	}

	return return_value;;
}

int Klotski::get_piece_id(char *src, int x, int y)
{
	return src[x + 1 + (y + 1) * (width+2)];
}

void Klotski::set_piece_id(char *src, int x, int y, int id)
{
	  src[x + 1 + (y + 1) * (width+2)] = id;
}


int Klotski::get_piece_nr(char *src, int x,int y)
{
	char c;
	int i=0, nr = 0;
	x++; y++;
	c = src[x + y * (width+2)];
	if(c=='-') return 23;
	if(c==' ') return 21;
	if(c=='.') return 20;

	nr += 1   * (src[(x - 1) + (y - 1) * (width+2)] == c);
	nr += 2   * (src[(x - 0) + (y - 1) * (width+2)] == c);
	nr += 4   * (src[(x + 1) + (y - 1) * (width+2)] == c);
	nr += 8   * (src[(x - 1) + (y - 0) * (width+2)] == c);
	nr += 16  * (src[(x + 1) + (y - 0) * (width+2)] == c);
	nr += 32  * (src[(x - 1) + (y + 1) * (width+2)] == c);
	nr += 64  * (src[(x - 0) + (y + 1) * (width+2)] == c);
	nr += 128 * (src[(x + 1) + (y + 1) * (width+2)] == c);
	while(nr != image_map[i] && image_map[i] != -1) i+=2;
	return image_map[i+1];
}

int Klotski::mapcmp(char *m1, char *m2)
{
	int x,y;
	for(y=0; y<height; y++)
		for(x=0; x<width; x++)
			if(get_piece_id(m1,x,y) != get_piece_id(m2,x,y))
				return 1;
	return 0;
}

void Klotski::copymap(char *dest,char *src)
{
	memcpy(dest,src,(width+2)*(height+1));
}

void Klotski::prepare_map(char *level)
{
	int x,y,i=0;
	static int first = 1;
	char tmp[32];
	char *p = level;

	if(p==NULL){
		if(first){
			p = \
			"aa  cc  ." \
			"aa ccc   " \
			"aa**     " \
			"ee*      " \
			"bb**     " \
			"bb ddd   " \
			"bb  dd  .";
			width = 9;
			height = 7;
			first = 0;
			QMessageBox::information(this, "Welcome","Welcome to Gnome Klotski");
		} else {
			return;
		}
	} else {
		while(i<16 && *p != '#')
			/* current_level[i++] = * */ p++;
		//current_level[i] = '\0'; 
		p++;

		i = 0; 
		while(i<16 && *p != '#') tmp[i++] = *p++;
		tmp[i] = '\0'; width = atoi(tmp); p++;

		i = 0; 
		while(i<16 && *p != '#') tmp[i++] = *p++;
		tmp[i] = '\0'; height = atoi(tmp); p++;

	}

	if(map){
		free(map);
		free(tmpmap);
		free(move_map);
		free(orig_map);
	}

	map = (char *) calloc(1,(width+2)*(height+2));
	tmpmap = (char *) calloc(1,(width+2)*(height+2));
	orig_map = (char *) calloc(1,(width+2)*(height+2));
	move_map = (char *) calloc(1,(width+2)*(height+2));

	if(p!=NULL)
		for(y=0; y<height; y++)
			for(x=0; x<width; x++)
				set_piece_id(map,x,y,*p++);
	copymap(orig_map,map);
}



void Klotski::about() 
{
	QMessageBox::about( this, "About klotski", 
			QString(APPNAME) +
			QString( "This is a basic port of the Lars Rydlinge klotski for Gnome.\nThe port was straightforward and done in less than 4 hours.\n\n"
	"(C) 1999 Lars Rydlinge <Lars.Rydlinge@HIG.SE> (for Gnome klotski)\n"
	"(C) 2000 Philippe Fremy <pfremy@chez.com>\n\n" ) );
}

void Klotski::aboutQt() 
{
	QMessageBox::aboutQt( this );
}

void Klotski::level(int i) 
{
       if (i < 1 || i > NUMBOARD) return;

	new_game(level_board[i-1]); 
}

main(int argc, char ** argv)
{
	QApplication a(argc, argv);
	Klotski * w = new Klotski();
	a.setMainWidget( w );
	w->show();
	a.exec();
}
