#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <qapplication.h>
#include <qmainwindow.h>
#include <qstatusbar.h>
#include <qmessagebox.h>
#include <qmenubar.h>
#include <qpainter.h>
#include <qlabel.h>


#define APPNAME "klotski"

#define TILE_SIZE 34
#define NUMBOARD  26


#ifndef KLOTSKI_H
#define KLOTSKI_H

class Board : public QWidget
{
	Q_OBJECT

public:
	Board(QWidget * parent) : QWidget( parent ) {}

signals:
	void sg_paintEvent( QPaintEvent * );
	void sg_pressEvent( QMouseEvent * );
	void sg_releaseEvent( QMouseEvent * );
	void sg_moveEvent( QMouseEvent * );

protected:
	void paintEvent( QPaintEvent * );
	void mousePressEvent( QMouseEvent * );
	void mouseReleaseEvent( QMouseEvent * );
	void mouseMoveEvent( QMouseEvent * );
};


class Klotski : public QMainWindow
{

	Q_OBJECT

public:
	Klotski();
	~Klotski();
	void init_menu();
	void load_image();
	void new_game( char * newmap=0L );
	void set_move(int m);
	void new_move();
	int game_over();
	void prepare_map(char *level);
	int movable(int id);
	int do_move_piece(int id, int dx, int dy);
	int check_valid_move(int id, int dx, int dy);
	int move_piece(int id,int x1,int y1,int x2,int y2);
	int get_piece_id(char *src, int x, int y);
	void set_piece_id(char *src, int x, int y, int id);
	int get_piece_nr(char *src, int x,int y);
	void gui_draw_pixmap(char *target, int x, int y);
	void print_map(char *src);
	int mapcmp(char *m1, char *m2);
	void copymap(char *dest,char *src);

public slots:
	/** update score with moves */
	void quit();
	void level(int);
	void about();
	void aboutQt();

protected slots:
	void space_mousePressEvent(QMouseEvent * event);
	void space_mouseReleaseEvent(QMouseEvent * event);
	void space_mouseMoveEvent(QMouseEvent * event);
	void space_paintEvent(QPaintEvent * pe);

protected:
	void closeEvent( QCloseEvent * ) { quit(); }

	QStatusBar * statusbar;
	Board * space;
	QLabel * move_lab;
	QPixmap * tiles_pixmap;
	QPixmap * buffer;

	char *map;
	char *tmpmap;
	char *move_map;
	char *orig_map;

	int height,width;
	int moves;

	int piece_id;
	int piece_x; 
	int piece_y; 
};

#endif
