{ return piece_id(map,x,y); }
{ return piece_id(orig_map,x,y)=='.'; }
{ return src[x + 1 + (y + 1) * (width+2)]; }
{ set_piece_id(map,x,y,id); }
{ quit(); }{ print(map); }
/* 
 *   Klotski: a port of Gnome Klotski to Qt
 *   Gnome Klotski written by Lars Rydlinge <lars.rydlinge@hig.se>
 *   Port to Qt by Philippe Fremy <pfremy@chez.com>
 * 
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "klotski.h"
#include "pieces.h"
#include "levels.h"

Map::Map()
: width(0), height(0), map(0), orig_map(0)
{ }

Map::~Map()
{
	if (map) delete map;
	if (orig_map) delete orig_map;
}

Map::Map(const Map & m)
{
	*this = m;
}

void Map::operator=(const Map & m)
{
	width = m.width;
	height = m.height;
	orig_map = new char[(width+2)*(height+1)];
	map = new char[(width+2)*(height+1)];
	memcpy(map,m.map,(width+2)*(height+1));	
	memcpy(orig_map,m.map,(width+2)*(height+1));	
}

void Map::prepare(char *level)
{
	int x,y,i=0;
	char tmp[32];
	char *p = level;

	while(i<16 && *p != '#')
		p++;
	p++;

	i = 0; 
	while(i<16 && *p != '#') tmp[i++] = *p++;
	tmp[i] = '\0'; width = atoi(tmp); p++;

	i = 0; 
	while(i<16 && *p != '#') tmp[i++] = *p++;
	tmp[i] = '\0'; height = atoi(tmp); p++;

	if (map) delete map;
	if (orig_map) delete orig_map;
	map = new char[(width+2)*(height+2)];
	orig_map = new char[(width+2)*(height+2)];

	if(p!=NULL)
		for(y=0; y<height; y++)
			for(x=0; x<width; x++)
				set_piece_id(x,y,*p++);
	memcpy(orig_map,map,(width+2)*(height+1));	
}

void Map::print(char * src)
{
	int x,y;
	for(y=0; y<height; y++) {
		for(x=0; x<width; x++)
			printf("%c", piece_id(src, x,y));
		printf("\n");
	}
}

int Map::game_over()
{
	if (! map || !orig_map) return 0;

	int x,y;
	for(y=0; y<height; y++)
		for(x=0; x<width; x++)
			if (piece_id(orig_map,x,y) == '.' && piece_id(x,y) != '*')
				return 0;
	return 1;
}

int Map::can_move_piece(int id, int dx, int dy)
{
	int x,y;

	for(y=0; y<height; y++)
		for(x=0; x<width; x++)
			if(piece_id(x,y) == id)
				if(!(piece_id(x+dx,y+dy) == ' ' ||
				     piece_id(x+dx,y+dy) == '.' ||
				     piece_id(x+dx,y+dy) == id ||
				   (id == '*' && piece_id(x+dx,y+dy) == '-'))){
					return 0;
				}
	return 1;
}

int Map::movable(int id)
{
	if(!(id == '#' || id == '.' || id == ' ' || id == '-'))
		return 1;
	return 0;
}

int Map::check_valid_move(int id,int x1,int y1,int x2,int y2)
{
	if (!movable(id)) 
		return 0;

	if(!((abs(y1-y2)==0 && abs(x1-x2)==1) || (abs(x1-x2)==0 && abs(y1-y2)==1)))
		return 0;

	return can_move_piece(id,x2-x1,y2-y1);
}

void Map::do_move_piece(int id, int dx, int dy)
{
	int x,y;
	char * tmpmap = new char[ (width+2)*(height+2)];	
	memcpy(tmpmap, map, (width+2)*(height+1) );

	/* Move pieces */
	for(y=0; y<height; y++)
		for(x=0; x<width; x++)
			if(piece_id(tmpmap, x,y) == id)
				set_piece_id(tmpmap,x,y,' ');

	for(y=0; y<height; y++)
		for(x=0; x<width; x++)
			if(piece_id(x,y) == id)
				set_piece_id(tmpmap,x+dx,y+dy,id);

	/* Preserve some from original map */
	for(y=0; y<height; y++)
		for(x=0; x<width; x++){
			if(piece_id(tmpmap,x,y)==' ' && piece_id(orig_map,x,y)=='.')
				set_piece_id(tmpmap,x,y,'.');
			if(piece_id(tmpmap,x,y)==' ' && piece_id(orig_map,x,y)=='-')
				set_piece_id(tmpmap,x,y,'-');
		}

	// Emit signal ?
	/* Paint changes */

	dirty_tiles.clear();
	for(y=0; y<height; y++)
		for(x=0; x<width; x++)
			if(piece_id(map,x,y) != piece_id(tmpmap,x,y) ||
				 piece_id(tmpmap,x,y) == id) {
					dirty_tiles.append(x);
					dirty_tiles.append(y);
				}

	memcpy(map, tmpmap, (width+2)*(height+1) );
       	delete tmpmap;
}

int Map::operator!=(Map m2)
{
	int x,y;
	for(y=0; y<height; y++)
		for(x=0; x<width; x++)
			if(piece_id(x,y) != m2.piece_id(x,y))
				return 1;
	return 0;
}

int Map::piece_nr(int x,int y)
{
	char c;
	int i=0, nr = 0;
	x++; y++;
	c = map[x + y * (width+2)];
	if(c=='-') return 23;
	if(c==' ') return 21;
	if(c=='.') return 20;

	nr += 1   * (map[(x - 1) + (y - 1) * (width+2)] == c);
	nr += 2   * (map[(x - 0) + (y - 1) * (width+2)] == c);
	nr += 4   * (map[(x + 1) + (y - 1) * (width+2)] == c);
	nr += 8   * (map[(x - 1) + (y - 0) * (width+2)] == c);
	nr += 16  * (map[(x + 1) + (y - 0) * (width+2)] == c);
	nr += 32  * (map[(x - 1) + (y + 1) * (width+2)] == c);
	nr += 64  * (map[(x - 0) + (y + 1) * (width+2)] == c);
	nr += 128 * (map[(x + 1) + (y + 1) * (width+2)] == c);
	while(nr != image_map[i] && image_map[i] != -1) i+=2;
	return image_map[i+1];
}

Board::Board(QWidget * parent) 
	: QWidget(parent)
{
	buffer = new QPixmap();
	buffer->resize(TILE_SIZE, TILE_SIZE);
	tiles_pixmap = new QPixmap( "gnotski.png" );
	map = 0L;
	piece_x = 0;
	piece_y = 0;
	piece_id = 0;
}

Board::~Board()
{
	delete tiles_pixmap;
	delete buffer;
}

void Board::load_map(Map * m)
{
	map = m;
	resize(m->width*TILE_SIZE, m->height*TILE_SIZE);
	setMinimumSize( size() );
}

void Board::paintEvent( QPaintEvent * pe)
{
	int x,y;
	for(y = 0; y < map->height; y++)
		for(x = 0; x < map->width; x++)
			gui_draw_pixmap(x, y);
}


void Board::mousePressEvent( QMouseEvent * me)
{
	piece_x = (int) me->x()/TILE_SIZE;
	piece_y = (int) me->y()/TILE_SIZE;
	piece_id = map->piece_id(piece_x,piece_y); 

	move_map = *map;
}


void Board::mouseReleaseEvent( QMouseEvent * me )
{
	if(*map != move_map) {
		emit new_move();
	}
}


void Board::mouseMoveEvent( QMouseEvent * me )
{
	int new_piece_x, new_piece_y;

	new_piece_x = (int) me->x()/TILE_SIZE;
	new_piece_y = (int) me->y()/TILE_SIZE;
	if(new_piece_x >= map->width || me->x() < 0 ||
	   new_piece_y >= map->height || me->y() < 0) 
		return;

	if(map->check_valid_move(piece_id,piece_x, piece_y, new_piece_x, new_piece_y)) {
		map->do_move_piece(piece_id,new_piece_x-piece_x, new_piece_y-piece_y);
		piece_x = new_piece_x; piece_y = new_piece_y;
		update_dirty_tiles();
	}
	return;
}

void Board::update_dirty_tiles()
{
	int x, y;
	QValueList<int>::Iterator it;
	for( it = map->dirty_tiles.begin(); it != map->dirty_tiles.end(); ++it) {
		x = *it;
		++it;
		y = *it;
		gui_draw_pixmap(x,y);
	}
}

void Board::gui_draw_pixmap(int x, int y) 
{
	int value;
	bitBlt( buffer, 0, 0, tiles_pixmap, 
		  map->piece_nr(x,y)*TILE_SIZE, 0, 
		  TILE_SIZE, TILE_SIZE, Qt::CopyROP);
  
	if(map->piece_id(x,y)=='*'){
		if(map->is_goal(x,y))
			value = 20;
		else
			value = 22;

		bitBlt( buffer, 8, 8, tiles_pixmap, value*TILE_SIZE+10,10,
		    8, 8, Qt::CopyROP);
	}

	QPainter p( this );
	p.drawPixmap( x*TILE_SIZE, y*TILE_SIZE, *buffer ); 
	p.end();
}

Klotski::Klotski()
	: QMainWindow()
{
	statusbar = new QStatusBar( this );
	move_lab = new QLabel( statusbar );
	set_move( 0 );
	move_lab->setMinimumSize( move_lab->sizeHint() );
	statusbar->addWidget( move_lab, 0, true );

	setCaption( APPNAME );
	init_menu();
	
	board = new Board(this);
	setCentralWidget( board );
	connect( board, SIGNAL(new_move()), this, SLOT(new_move()));

	QMessageBox::information(this, "Welcome","Welcome to Gnome Klotski");
	new_game();

}


void Klotski::init_menu() {
	QPopupMenu * level_1_menu = new QPopupMenu( this );
	QPopupMenu * level_2_menu = new QPopupMenu( this );
	QPopupMenu * level_3_menu = new QPopupMenu( this );
	level_1_menu->insertItem( "1", 1);
	level_1_menu->insertItem( "2", 2);
	level_1_menu->insertItem( "3", 3);
	level_1_menu->insertItem( "4", 4);
	level_1_menu->insertItem( "5", 5);
	level_1_menu->insertItem( "6", 6);
	level_1_menu->insertItem( "7", 7);

	level_2_menu->insertItem( "11", 11);
	level_2_menu->insertItem( "12", 12);
	level_2_menu->insertItem( "13", 13);
	level_2_menu->insertItem( "14", 14);
	level_2_menu->insertItem( "15", 15);
	level_2_menu->insertItem( "16", 16);
	level_2_menu->insertItem( "17", 17);

	level_3_menu->insertItem( "21", 21);
	level_3_menu->insertItem( "22", 22);
	level_3_menu->insertItem( "23", 23);
	level_3_menu->insertItem( "24", 24);
	level_3_menu->insertItem( "25", 25);
	level_3_menu->insertItem( "26", 26);

	QPopupMenu * game_menu = new QPopupMenu();
	game_menu->insertItem( "Novice", level_1_menu);
	game_menu->insertItem( "Medium", level_2_menu);
	game_menu->insertItem( "Advanced", level_3_menu);

	QPopupMenu * file_menu = new QPopupMenu();
	file_menu->insertItem( "quit", this, SLOT( quit() ));

	QPopupMenu * help_menu = new QPopupMenu( this );
	help_menu->insertItem( "About klotski", this, SLOT( about() ));
	help_menu->insertItem( "About Qt", this, SLOT( aboutQt() ));

	QMenuBar * main_menu = menuBar();
	main_menu->insertItem( "File", file_menu );
	main_menu->insertItem( "Game", game_menu );
	main_menu->insertItem( "Help", help_menu );

	connect( level_1_menu, SIGNAL( activated( int )),
		this, SLOT( level( int ) ));
	connect( level_2_menu, SIGNAL( activated( int )),
		this, SLOT( level( int ) ));
	connect( level_3_menu, SIGNAL( activated( int )),
		this, SLOT( level( int ) ));
}




void Klotski::quit()
{
	if (moves) {
		if (0 != QMessageBox::warning( this, "Quit", "Really quit ?", "Ok", "Cancel"))
			return;
	}

	qApp->quit();
}

void Klotski::new_game( char * newmap )
{
	prepare_map( newmap );
	board->load_map( &map);
	resize( sizeHint() );
	set_move(0);
}

void Klotski::set_move(int x) 
{
	moves = x-1;
	new_move();
}

void Klotski::new_move() {
	moves++;
	QString s;
	s.sprintf("Moves : %4d", moves);
	move_lab->setText( s );
	if (map.game_over()) {
		QMessageBox::information( this, "", "Level completed. Well done."); 
	}
}



void Klotski::prepare_map(char *level)
{
	if(level==NULL){
		level = \
                "Klotski#9#7#" \
		"aa  cc  ." \
		"aa ccc   " \
		"aa**     " \
		"ee*      " \
		"bb**     " \
		"bb ddd   " \
		"bb  dd  .";
	}
	map.prepare( level );
}



void Klotski::about() 
{
	QMessageBox::about( this, "About klotski", 
			QString(APPNAME) +
			QString( "This is a port of the Lars Rydlinge klotski for Gnome.\nThe port was straightforward and done in less than 4 hours.\n\n"
	"(C) 1999 Lars Rydlinge <Lars.Rydlinge@HIG.SE> (for Gnome klotski)\n"
	"(C) 2000 Philippe Fremy <pfremy@chez.com>\n\n" ) );
}

void Klotski::aboutQt() 
{
	QMessageBox::aboutQt( this );
}

void Klotski::level(int i) 
{
       if (i < 1 || i > NUMBOARD) return;
	new_game(level_board[i-1]); 
}

main(int argc, char ** argv)
{
	QApplication a(argc, argv);
	Klotski * w = new Klotski();
	a.setMainWidget( w );
	w->show();
	a.exec();
}
