/* 
 *   Klotski: a port of Gnome Klotski to Qt
 *   Gnome Klotski written by Lars Rydlinge <lars.rydlinge@hig.se>
 *   Port to Qt by Philippe Fremy <pfremy@chez.com>
 * 
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <qapplication.h>
#include <qmainwindow.h>
#include <qstatusbar.h>
#include <qmessagebox.h>
#include <qmenubar.h>
#include <qpainter.h>
#include <qlabel.h>
#include <qvaluelist.h>


#define APPNAME "klotski"

#define TILE_SIZE 34
#define NUMBOARD  26


#ifndef KLOTSKI_H
#define KLOTSKI_H


class Map
{
public:
	Map();
	~Map();
	Map(const Map & m);
	void operator=(const Map & m);

	void prepare(char *level);

	int piece_id(int x, int y)

	void set_piece_id(int x, int y, int id)

	int piece_nr(int x,int y);
	void print()


	int check_valid_move(int id, int x1, int y1, int x2, int y2);
	int movable(int id);
	int can_move_piece(int id, int dx, int dy);
	void do_move_piece(int id, int dx, int dy);

	int game_over();
	bool is_goal(int x, int y)

	int operator!=(Map m2);

	QValueList<int> dirty_tiles;
	int width, height;

protected:
	void print(char * src);

	int piece_id(char * src, int x, int y)

	void set_piece_id(char * src, int x, int y, int id)
	{ src[x + 1 + (y + 1) * (width+2)] = id; }

	char * map;
	char * orig_map;

};



class Board : public QWidget
{
	Q_OBJECT

public:
	Board(QWidget * parent);
	~Board();

	void load_map(Map * m);
signals:
	void new_move();

protected:
	void paintEvent( QPaintEvent * );
	void mousePressEvent( QMouseEvent * );
	void mouseReleaseEvent( QMouseEvent * );
	void mouseMoveEvent( QMouseEvent * );

	void gui_draw_pixmap(int x, int y);
	void update_dirty_tiles();

	int piece_x, piece_y;
	int piece_id;

	Map * map;
	Map move_map;
	QPixmap * tiles_pixmap;
	QPixmap * buffer;
};


class Klotski : public QMainWindow
{

	Q_OBJECT

public:
	Klotski();
	void init_menu();
	void new_game( char * newmap=0L );
	void set_move(int m);
	void prepare_map(char *level);

public slots:
	/** update score with moves */
	void new_move();
	void quit();
	void level(int);
	void about();
	void aboutQt();

protected:
	void closeEvent( QCloseEvent * )

	QStatusBar * statusbar;
	Board * board;
	QLabel * move_lab;

	Map map;
	int moves;
};

#endif
