
#  Klotski: a port of Gnome Klotski to Qt
#  Gnome Klotski written by Lars Rydlinge <lars.rydlinge@hig.se>
#  Port to Qt by Philippe Fremy <pfremy@chez.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import copy, sys, string

from qt import *
from pieces import image_map
from levels import level_board

APPNAME = "klotski"
TILE_SIZE = 34
NUMBOARD = 26

class Map:
	def __init__(self):
		self.width = 0
		self.height = 0
		self.map = None
		self.orig_map = None
		self.dirty_tiles = []

	def piece_id(self,x,y,src=None):
		if not src: src = self.map
		return src[x + 1 + (y + 1) * (self.width+2)]

	def set_piece_id(self,x, y, id, src=None):
		if not src: src=self.map
		src[x + 1 + (y + 1) * (self.width+2)] = id

	def is_goal(self,x, y):
		return self.piece_id(x,y,self.orig_map) == '.'

	def prepare(self, level):
		l = string.split( level, '+' )
		if len(l) < 4: return

		self.width = int(l[1])
		self.height = int(l[2])
		self.map = [ ' ' ] * ((self.width+2)*(self.height+2))

		l = list(l[3])
		i=0
		for y in range(self.height):
			for x in range(self.width):
					self.set_piece_id(x,y,l[i])
					i=i+1
		self.orig_map = copy.deepcopy( self.map )

	def printmap(self,src=None):
		if not src: src = self.map
		for y in range(self.height):
			for x in range(self.width):
				print self.piece_id(x,y,src),
			print

	def game_over(self):
		if not self.map or not self.orig_map: return 0

		for y in range(self.height):
			for x in range(self.width):
				if self.piece_id(x,y,self.orig_map) == '.' and self.piece_id(x,y) != '*':
					return 0
		return 1

	def can_move_piece(self,id, dx, dy):
		for y in range(self.height):
			for x in range(self.width):
				if self.piece_id(x,y) == id :
					if not (self.piece_id(x+dx,y+dy) == ' ' or
					     self.piece_id(x+dx,y+dy) == '.' or
					     self.piece_id(x+dx,y+dy) == id or
					   (id == '*' and self.piece_id(x+dx,y+dy) == '-')):
						return 0
		return 1

	def movable(self,id):
		if not (id == '#' or id == '.' or id == ' ' or id == '-'):
			return 1
		return 0

	def check_valid_move(self,id,x1,y1,x2,y2):
		if not self.movable(id):
			return 0

		if not ((abs(y1-y2)==0 and abs(x1-x2)==1) or (abs(x1-x2)==0 and abs(y1-y2)==1)):
			return 0

		return self.can_move_piece(id,x2-x1,y2-y1)

	def do_move_piece(self,id, dx, dy):
		tmpmap = copy.deepcopy( self.map )

		# Move pieces
		for y in range(self.height):
			for x in range(self.width):
				if self.piece_id(x,y,tmpmap) == id:
					self.set_piece_id(x,y,' ',tmpmap)

		for y in range(self.height):
			for x in range(self.width):
				if self.piece_id(x,y) == id:
					self.set_piece_id(x+dx,y+dy,id,tmpmap)

		# Preserve some from original map
		for y in range(self.height):
			for x in range(self.width):
				if self.piece_id(x,y,tmpmap)==' ' and self.piece_id(x,y,self.orig_map)=='.':
					self.set_piece_id(x,y,'.',tmpmap)
				if self.piece_id(x,y,tmpmap)==' ' and self.piece_id(x,y,self.orig_map)=='-' :
					self.set_piece_id(x,y,'-',tmpmap)

		# Paint changes
		self.dirty_tiles = []
		for y in range(self.height):
			for x in range(self.width):
				if self.piece_id(x,y) != self.piece_id(x,y,tmpmap) or self.piece_id(x,y,tmpmap) == id :
						self.dirty_tiles.append((x,y))

		self.map = tmpmap

	def piece_nr(self,x,y):
		x=x+1 
		y=y+1
		c = self.map[x + y * (self.width+2)]
		if c=='-': return 23
		if c==' ': return 21
		if c=='.': return 20

		nr = 0
		nr = nr + 1   * (self.map[(x - 1) + (y - 1) * (self.width+2)] == c)
		nr = nr + 2   * (self.map[(x - 0) + (y - 1) * (self.width+2)] == c)
		nr = nr + 4   * (self.map[(x + 1) + (y - 1) * (self.width+2)] == c)
		nr = nr + 8   * (self.map[(x - 1) + (y - 0) * (self.width+2)] == c)
		nr = nr + 16  * (self.map[(x + 1) + (y - 0) * (self.width+2)] == c)
		nr = nr + 32  * (self.map[(x - 1) + (y + 1) * (self.width+2)] == c)
		nr = nr + 64  * (self.map[(x - 0) + (y + 1) * (self.width+2)] == c)
		nr = nr + 128 * (self.map[(x + 1) + (y + 1) * (self.width+2)] == c)

		i=0
		while nr != image_map[i] and image_map[i] != -1 : 
			i = i+2
		return image_map[i+1]

	def __cmp__(self, m2):
		for y in range(self.height):
			for x in range(self.width):
				if self.piece_id(x,y) != m2.piece_id(x,y) :
					return 1
		return 0


class Board (QWidget) :
	def __init__(self, parent):
		QWidget.__init__(self,parent)
		self.buffer = QPixmap()
		self.buffer.resize(TILE_SIZE, TILE_SIZE)
		self.tiles_pixmap = QPixmap( "gnotski.png" )
		self.map = None
		self.move_map = None
		self.piece_x = 0
		self.piece_y = 0
		self.piece_id = None

	def load_map(self,m):
		self.map = m;
		self.resize(m.width*TILE_SIZE, m.height*TILE_SIZE)
		self.setMinimumSize( self.size() )

	def paintEvent(self,pe):
		for y in range( self.map.height) :
			for x in range( self.map.width) :
				self.gui_draw_pixmap(x, y)

	def mousePressEvent(self,me):
		self.piece_x = int( me.x()/TILE_SIZE )
		self.piece_y = int( me.y()/TILE_SIZE )
		self.piece_id = self.map.piece_id(self.piece_x,self.piece_y)

		self.move_map = copy.deepcopy( self.map )

	def mouseReleaseEvent( self,me ):
		if self.map != self.move_map :
			self.emit(PYSIGNAL("new_move()"), ())

	def mouseMoveEvent( self, me ):
		new_piece_x = int( me.x()/TILE_SIZE )
		new_piece_y = int( me.y()/TILE_SIZE )
		if new_piece_x >= self.map.width or me.x() < 0 \
		   or new_piece_y >= self.map.height or me.y() < 0 :
			return

		if self.map.check_valid_move(self.piece_id, self.piece_x, self.piece_y, 
		    new_piece_x, new_piece_y) :
			self.map.do_move_piece(self.piece_id,new_piece_x-self.piece_x, new_piece_y-self.piece_y)
			self.piece_x = new_piece_x
			self.piece_y = new_piece_y
			self.update_dirty_tiles()
		return

	def update_dirty_tiles(self):
		for it in self.map.dirty_tiles:
			self.gui_draw_pixmap(it[0],it[1])

	def gui_draw_pixmap(self,x, y) :
		bitBlt( self.buffer, 0, 0, self.tiles_pixmap, 
			  self.map.piece_nr(x,y)*TILE_SIZE, 0, 
			  TILE_SIZE, TILE_SIZE, Qt.CopyROP)
	  
		if self.map.piece_id(x,y)=='*' :
			if self.map.is_goal(x,y) :
				value = 20
			else:
				value = 22

			bitBlt( self.buffer, 8, 8, self.tiles_pixmap, value*TILE_SIZE+10,10,
			    8, 8, Qt.CopyROP)

		p = QPainter( self )
		p.drawPixmap( x*TILE_SIZE, y*TILE_SIZE, self.buffer ); 
		p.end()


class Klotski(QMainWindow):

	def __init__(self):
		QMainWindow.__init__(self)
		self.map = Map()	
		self.board = Board(self)

		statusbar = QStatusBar( self )
		self.move_lab = QLabel( statusbar )
		self.set_move( 0 )
		self.move_lab.setMinimumSize( self.move_lab.sizeHint() )
		statusbar.addWidget( self.move_lab, 0, 1 )

		self.setCaption( APPNAME )
		self.init_menu()
	
		self.setCentralWidget( self.board )
		QObject.connect( self.board, PYSIGNAL("new_move()"), self.new_move )

		QMessageBox.information(self, "Welcome","Welcome to Gnome Klotski")
		self.new_game()


	def init_menu(self):
		level_1_menu = QPopupMenu( self )
		level_2_menu = QPopupMenu( self )
		level_3_menu = QPopupMenu( self )

		level_1_menu.insertItem( "1", 1)
		level_1_menu.insertItem( "2", 2)
		level_1_menu.insertItem( "3", 3)
		level_1_menu.insertItem( "4", 4)
		level_1_menu.insertItem( "5", 5)
		level_1_menu.insertItem( "6", 6)
		level_1_menu.insertItem( "7", 7)

		level_2_menu.insertItem( "11", 11)
		level_2_menu.insertItem( "12", 12)
		level_2_menu.insertItem( "13", 13)
		level_2_menu.insertItem( "14", 14)
		level_2_menu.insertItem( "15", 15)
		level_2_menu.insertItem( "16", 16)
		level_2_menu.insertItem( "17", 17)

		level_3_menu.insertItem( "21", 21)
		level_3_menu.insertItem( "22", 22)
		level_3_menu.insertItem( "23", 23)
		level_3_menu.insertItem( "24", 24)
		level_3_menu.insertItem( "25", 25)
		level_3_menu.insertItem( "26", 26)

		game_menu = QPopupMenu(self)
		game_menu.insertItem( "Novice", level_1_menu)
		game_menu.insertItem( "Medium", level_2_menu)
		game_menu.insertItem( "Advanced", level_3_menu)

		file_menu = QPopupMenu(self)
		file_menu.insertItem( "quit", self.quit )

		help_menu = QPopupMenu(self)
		help_menu.insertItem( "About klotski", self.about )
		help_menu.insertItem( "About Qt", self.aboutQt )

		main_menu = QMenuBar(self)
		main_menu.insertItem( "File", file_menu )
		main_menu.insertItem( "Game", game_menu )
		main_menu.insertItem( "Help", help_menu )
		main_menu.show()

		QObject.connect( level_1_menu, SIGNAL("activated(int)"), self.level)
		QObject.connect( level_2_menu, SIGNAL("activated(int)"), self.level)
		QObject.connect( level_3_menu, SIGNAL("activated(int)"), self.level)

	def quit(self):
		if self.moves :
			if 0 != QMessageBox.warning( self, "Quit", "Really quit ?", "Ok", "Cancel"):
				return

		qApp.quit()

	def new_game( self, newmap=None ):
		print "new game"
		self.prepare_map( newmap )
		self.board.load_map( self.map)
		self.resize( 0, 0 )
		#self.resize( self.sizeHint() )
		self.set_move(0)
		self.repaint()

	def set_move(self,x) :
		self.moves = x-1
		self.new_move()

	def new_move(self):
		self.moves=self.moves+1;
		self.move_lab.setText( "Moves : %4d" % self.moves )
		if self.map.game_over() :
			QMessageBox.information( self, "", "Level completed. Well done.") 

	def prepare_map(self,level=None):
		if level==None :
			level = \
			"Klotski+9+7+" \
			"aa  cc  ." \
			"aa ccc   " \
			"aa**     " \
			"ee*      " \
			"bb**     " \
			"bb ddd   " \
			"bb  dd  ."

		self.map.prepare( level )

	def about(self) :
		QMessageBox.about( self, "About klotski", 
				APPNAME + "This is a port of the Lars Rydlinge klotski for Gnome.\nThe port was straightforward and done in less than 4 hours.\n\n"
		"(C) 1999 Lars Rydlinge <Lars.Rydlinge@HIG.SE> (for Gnome klotski)\n"
		"(C) 2000 Philippe Fremy <pfremy@chez.com>\n\n" )

	def aboutQt(self):
		QMessageBox.aboutQt(self, "About QT")

	def level(self, i) :
		if i < 1 or i > NUMBOARD : return
		self.new_game(level_board[i-1])

# main
a = QApplication(sys.argv)
w = Klotski()
a.setMainWidget( w )
w.show()
a.exec_loop()
del w
